export interface IEscapeMethod {
    (text: string): string;
}
export declare const Escape_HtmlEntities: IEscapeMethod;
export interface IEmitterArgumentDescription {
    what: string;
    can_be_undefined: boolean;
    possible_values?: string[];
}
export declare type IEmitterArgumentsDescriptions = IEmitterArgumentDescription[];
export declare type TEmitterArguments = string[];
export declare const emitter_arguments_count_max: number;
export interface IEmissionMethod {
    (chunks: TChunks, emitters: IEmitters, locals?: TLocals, emitter_arguments?: TEmitterArguments): void;
}
export interface IEmitterDescription {
    what: string;
    arguments_descriptions: IEmitterArgumentsDescriptions;
    is_pure: boolean;
}
export interface IEmitter extends IEmitterDescription {
    Emit: IEmissionMethod;
}
export interface IBlockEmitter extends IEmitterDescription {
    Emit_Prologue: IEmissionMethod;
    Emit_Epilogue: IEmissionMethod;
}
export interface IEmitters {
    [index: string]: IEmitter | IBlockEmitter;
}
export declare type TChunks = string[];
export declare type TLocals = any;
export declare type TCommandsSequence = (string | number)[];
export declare class TValidationErrorException extends Error {
    private m_problem_command_index;
    private m_problem_emitter_name;
    constructor(description: string, problem_command_index?: number, problem_emitter_name?: string);
    readonly description: string;
    readonly problem_command_index: number;
    readonly problem_emitter_name: string;
}
export declare const Validate: (emitters: IEmitters) => void;
export interface ICompilationOptions {
    spaces_per_ident_level: number;
    raw_text_is_allowed: boolean;
    escape_method: IEscapeMethod;
}
export declare class TCompilationErrorException extends Error {
    private m_line;
    private m_line_index;
    private m_line_char_index;
    constructor(description: string, line_index: number, line_begin_char_index: number, problem_char_index: number, template: string);
    private static Extract_Line(line_begin_char_index, problem_char_index, template);
    readonly description: string;
    readonly line: string;
    readonly line_index: number;
    readonly line_char_index: number;
}
export declare const Compile: (template: string, options?: ICompilationOptions) => (string | number)[];
export declare const Advance: (emitters: IEmitters) => (string | number)[];
export declare const Emit: (chunks: string[], emitters: IEmitters, locals?: any) => void;
export declare const Render: (emitters: IEmitters, locals?: any) => string;
export declare class TRenderer implements IEmitter {
    private static TRenderer_s_arguments_descriptions;
    private m_commands_sequence;
    constructor(command_sequence: TCommandsSequence);
    Validate(emitters: IEmitters): void;
    Advance(emitters: IEmitters): void;
    Render(emitters: IEmitters, locals?: TLocals): string;
    readonly what: string;
    readonly arguments_descriptions: IEmitterArgumentDescription[];
    readonly is_pure: boolean;
    Emit(chunks: TChunks, emitters: IEmitters, locals?: TLocals, emitter_arguments?: TEmitterArguments): void;
}
