"use strict";
exports.Escape_HtmlEntities = function (text) {
    return (text
        .split('&').join('&amp;')
        .split('"').join('&quot;')
        .split('<').join('&lt;')
        .split('>').join('&gt;')
        .split('\'').join('&#x27;')
        .split('/').join('&#x2F;'));
};
exports.emitter_arguments_count_max = 8;
class TValidationErrorException extends Error {
    constructor(description, problem_command_index, problem_emitter_name) {
        super(description);
        this.m_problem_command_index = problem_command_index;
        this.m_problem_emitter_name = problem_emitter_name;
    }
    get description() { return (this.message); }
    get problem_command_index() { return (this.m_problem_command_index); }
    get problem_emitter_name() { return (this.m_problem_emitter_name); }
}
exports.TValidationErrorException = TValidationErrorException;
function On_ValidationError(description, problem_command_index, problem_emitter_name) {
    throw (new TValidationErrorException(description, problem_command_index, problem_emitter_name));
}
function Validate_Emitters(emitters) {
    for (let emitter_name in emitters) {
        const emitter_name_chars_count = emitter_name.length;
        if (emitter_name_chars_count <= 0) {
            On_ValidationError('emitter name is not valid', undefined, emitter_name);
        }
        else {
            let emitter_name_char_index = 0;
            do {
                ++emitter_name_char_index;
            } while (emitter_name_chars_count !== emitter_name_char_index);
        }
        const emitter = emitters[emitter_name];
        const right_slash_char_code = 47;
        if (right_slash_char_code !== emitter_name.charCodeAt(emitter_name.length - 1)) {
            if (('function' !== typeof (emitter.Emit_Prologue))
                ||
                    ('function' !== typeof (emitter.Emit_Epilogue))) {
                On_ValidationError('emitter does not implement IBlockEmitter interface', undefined, emitter_name);
            }
        }
        else {
            if ('function' !== typeof (emitter.Emit)) {
                On_ValidationError('emitter does not implement IEmitter interface', undefined, emitter_name);
            }
        }
        if (exports.emitter_arguments_count_max < emitter.arguments_descriptions.length) {
            On_ValidationError('emitter requires too many arguments', undefined, emitter_name);
        }
    }
}
function Validate_CommandsSequence(emitters) {
    const commands_sequence = this;
    const commands_count = commands_sequence.length;
    let block_emitters_commands_start_indexes = [];
    let last_command_was_text_chunk = false;
    let command_index = 0;
    while (commands_count !== command_index) {
        const command = commands_sequence[command_index];
        switch (typeof command) {
            case 'string':
                {
                    if (last_command_was_text_chunk) {
                        On_ValidationError('consecutive document chunk commands', command_index);
                    }
                    const string_command = command;
                    if (string_command.length <= 0) {
                        On_ValidationError('document chunk command value is not valid', command_index);
                    }
                    command_index += 1;
                    last_command_was_text_chunk = true;
                    break;
                }
            case 'number':
                {
                    const number_command = command;
                    if (((-1 - exports.emitter_arguments_count_max) <= number_command) && (number_command <= -1)) {
                        const emitter_commands_start_index = command_index;
                        const emitter_arguments_count = -1 - number_command;
                        const emitter_arguments_commands_begin_index = emitter_commands_start_index + 1;
                        const emitter_arguments_commands_end_index = emitter_arguments_commands_begin_index + emitter_arguments_count;
                        const emitter_name_command_index = emitter_arguments_commands_end_index;
                        if (commands_count <= emitter_name_command_index) {
                            On_ValidationError('emitter arguments count command value is not valid', command_index);
                        }
                        const emitter_name = commands_sequence[emitter_name_command_index];
                        if ('string' !== typeof emitter_name) {
                            On_ValidationError('emitter name command type is not valid', emitter_name_command_index);
                        }
                        if (emitter_name.length <= 0) {
                            On_ValidationError('emitter name command value is not valid', emitter_name_command_index);
                        }
                        const right_slash_char_code = 47;
                        const block_opens = (right_slash_char_code !== emitter_name.charCodeAt(emitter_name.length - 1));
                        const emitter = emitters[emitter_name];
                        if (undefined === emitter) {
                            On_ValidationError('emitter not found', emitter_name_command_index, emitter_name);
                        }
                        {
                            if (emitter.arguments_descriptions.length < emitter_arguments_count) {
                                On_ValidationError('emitter arguments count command value is out of bounds', emitter_commands_start_index, emitter_name);
                            }
                            let emitter_argument_command_index = emitter_arguments_commands_begin_index;
                            let emitter_argument_index = 0;
                            while (emitter.arguments_descriptions.length != emitter_argument_index) {
                                if ((false === emitter.arguments_descriptions[emitter_argument_index].can_be_undefined)
                                    &&
                                        ((emitter_arguments_commands_end_index === emitter_argument_command_index)
                                            ||
                                                (undefined === commands_sequence[emitter_argument_command_index]))) {
                                    On_ValidationError('emitter argument can not be undefined', emitter_argument_command_index, emitter_name);
                                }
                                if (emitter_arguments_commands_end_index !== emitter_argument_command_index) {
                                    switch (typeof commands_sequence[emitter_argument_command_index]) {
                                        case 'undefined':
                                        case 'string':
                                            {
                                                break;
                                            }
                                        default:
                                            {
                                                On_ValidationError('emitter argument command type is not valid', emitter_argument_command_index, emitter_name);
                                            }
                                    }
                                    ++emitter_argument_command_index;
                                }
                                ++emitter_argument_index;
                            }
                        }
                        if (block_opens) {
                            block_emitters_commands_start_indexes.push(emitter_commands_start_index);
                        }
                        command_index += 2 + emitter_arguments_count;
                    }
                    else if ((0 <= number_command) && (number_command < commands_count)) {
                        const closing_emitter_arguments_start_index = number_command;
                        if (0 === block_emitters_commands_start_indexes.length) {
                            On_ValidationError('block close numeric command is not expected', command_index);
                        }
                        const expected_closing_emitter_arguments_start_index = block_emitters_commands_start_indexes.pop();
                        if (expected_closing_emitter_arguments_start_index !== closing_emitter_arguments_start_index) {
                            On_ValidationError('block close numeric command value is not valid', command_index);
                        }
                        command_index += 1;
                    }
                    else {
                        On_ValidationError('numeric command value is not valid', command_index);
                    }
                    last_command_was_text_chunk = false;
                    break;
                }
            default:
                {
                    On_ValidationError('command type is not valid', command_index);
                }
        }
    }
    if (0 !== block_emitters_commands_start_indexes.length) {
        On_ValidationError('not all the blocks are closed', command_index);
    }
}
exports.Validate = function (emitters) {
    if (undefined !== this) {
        Validate_CommandsSequence.apply(this, arguments);
    }
    else {
        Validate_Emitters.apply(this, arguments);
    }
};
function Is_Whitespace(char_code) {
    return ((32 === char_code) || (9 === char_code));
}
function Is_Numeric(char_code) {
    return ((48 <= char_code) && (char_code <= 57));
}
function Is_Uppercase(char_code) {
    return ((65 <= char_code) && (char_code <= 90));
}
function Is_Lowercase(char_code) {
    return ((97 <= char_code) && (char_code <= 122));
}
function Is_NameSafe(char_code) {
    return (Is_Numeric(char_code) || Is_Uppercase(char_code) || Is_Lowercase(char_code) || (95 === char_code));
}
class TCompilationErrorException extends Error {
    constructor(description, line_index, line_begin_char_index, problem_char_index, template) {
        super(description);
        this.m_line = TCompilationErrorException.Extract_Line(line_begin_char_index, problem_char_index, template);
        this.m_line_index = line_index;
        this.m_line_char_index = problem_char_index - line_begin_char_index;
    }
    static Extract_Line(line_begin_char_index, problem_char_index, template) {
        let line_end_char_index = problem_char_index;
        while ((template.length !== line_end_char_index)
            &&
                (13 !== template.charCodeAt(line_end_char_index))
            &&
                (10 !== template.charCodeAt(line_end_char_index))) {
            ++line_end_char_index;
        }
        return (template.substring(line_begin_char_index, line_end_char_index));
    }
    get description() { return (this.message); }
    get line() { return (this.m_line); }
    get line_index() { return (this.m_line_index); }
    get line_char_index() { return (this.m_line_char_index); }
}
exports.TCompilationErrorException = TCompilationErrorException;
class TCompilationContext {
    constructor(template, options) {
        this.m_template = template;
        this.m_current_char_index = 0;
        this.m_current_line_index = 0;
        this.m_current_line_first_char_index = 0;
        this.m_current_depth = 0;
        this.m_last_block_depth = 0;
        this.m_last_block_closed = true;
        this.m_opened_blocks_commands_starts_indexes = [];
        this.m_text_chunks_accumulator = [];
        this.m_options = (undefined !== options) ? options : TCompilationContext.s_default_options;
        this.m_commands_sequence = [];
        while (template.length !== this.m_current_char_index) {
            this.Parse_Indentation();
            this.Parse_Line();
            this.Skip_Eol();
        }
        this.Flush_AccumulatedTextChunks();
        while (0 !== this.m_opened_blocks_commands_starts_indexes.length) {
            this.m_commands_sequence.push(this.m_opened_blocks_commands_starts_indexes.pop());
        }
    }
    get compiled_commands() {
        return (this.m_commands_sequence);
    }
    Throw(description) {
        throw (new TCompilationErrorException(description, this.m_current_line_index, this.m_current_line_first_char_index, this.m_current_char_index, this.m_template));
    }
    Flush_AccumulatedTextChunks() {
        if (0 !== this.m_text_chunks_accumulator.length) {
            this.m_commands_sequence.push(this.m_text_chunks_accumulator.join(''));
        }
        this.m_text_chunks_accumulator.length = 0;
    }
    Prepare_To_AppendCommand() {
        let depth_delta = this.m_current_depth - this.m_last_block_depth;
        if (1 < depth_delta) {
            this.Throw('Block depth increased by more than 1 level.');
        }
        else if ((1 === depth_delta) && this.m_last_block_closed) {
            this.Throw('Block depth increase is prohibited.');
        }
        else if (depth_delta < 0) {
            this.Flush_AccumulatedTextChunks();
            do {
                this.m_commands_sequence.push(this.m_opened_blocks_commands_starts_indexes.pop());
                ++depth_delta;
            } while (depth_delta !== 0);
        }
        if ((0 === depth_delta) && (!this.m_last_block_closed)) {
            this.Flush_AccumulatedTextChunks();
            this.m_commands_sequence.push(this.m_opened_blocks_commands_starts_indexes.pop());
        }
        this.m_last_block_depth = this.m_current_depth;
    }
    Parse_Indentation() {
        const indentation_begin_char_index = this.m_current_char_index;
        const first_indentation_char_code = this.m_template.charCodeAt(this.m_current_char_index);
        if (9 === first_indentation_char_code) {
            for (;;) {
                ++this.m_current_char_index;
                if (this.m_template.length === this.m_current_char_index) {
                    break;
                }
                const char_code = this.m_template.charCodeAt(this.m_current_char_index);
                if (32 === char_code) {
                    this.Throw('unexpected space at tab indentation');
                }
                else if (9 !== char_code) {
                    break;
                }
            }
            this.m_current_depth = (this.m_current_char_index - indentation_begin_char_index);
        }
        else if (32 === first_indentation_char_code) {
            for (;;) {
                ++this.m_current_char_index;
                if (this.m_template.length === this.m_current_char_index) {
                    break;
                }
                const char_code = this.m_template.charCodeAt(this.m_current_char_index);
                if (9 === char_code) {
                    this.Throw('unexpected tab at space indentation');
                }
                else if (32 !== char_code) {
                    break;
                }
            }
            this.m_current_depth = (this.m_current_char_index - indentation_begin_char_index);
            if (0 !== (this.m_current_depth % this.m_options.spaces_per_ident_level)) {
                this.Throw('amount of spaces in space indentation is not multiple of ' + this.m_options.spaces_per_ident_level.toString());
            }
            this.m_current_depth = (this.m_current_depth / this.m_options.spaces_per_ident_level);
        }
        else {
            this.m_current_depth = 0;
        }
    }
    Parse_Line() {
        if (this.m_template.length !== this.m_current_char_index) {
            const char_code = this.m_template.charCodeAt(this.m_current_char_index);
            if (Is_NameSafe(char_code)) {
                this.Parse_Emitter();
            }
            else if (92 === char_code) {
                this.Prepare_To_AppendCommand();
                this.m_text_chunks_accumulator.push(this.m_options.escape_method(this.Parse_TextLine()));
                this.m_last_block_closed = true;
            }
            else if (61 === char_code) {
                if (this.m_options.raw_text_is_allowed) {
                    this.Prepare_To_AppendCommand();
                    this.m_text_chunks_accumulator.push(this.Parse_TextLine());
                    this.m_last_block_closed = true;
                }
                else {
                    this.Throw('Raw text processing is prohibited');
                }
            }
            else if (45 === char_code) {
                this.Skip_SingleLineComment();
            }
            else if (13 === char_code) {
            }
            else if (10 === char_code) {
            }
            else {
                this.Throw('Unexpected char');
            }
        }
    }
    Skip_OptionalWhitespace() {
        while ((this.m_template.length !== this.m_current_char_index)
            &&
                Is_Whitespace(this.m_template.charCodeAt(this.m_current_char_index))) {
            ++this.m_current_char_index;
        }
    }
    Skip_Eol() {
        if (this.m_template.length !== this.m_current_char_index) {
            const char_code = this.m_template.charCodeAt(this.m_current_char_index);
            if (13 === char_code) {
                ++this.m_current_char_index;
                if (this.m_template.length === this.m_current_char_index) {
                    this.Throw('Unexpected EOF at \r\n EOL sequence.');
                }
                if (10 !== this.m_template.charCodeAt(this.m_current_char_index)) {
                    this.Throw('Unexpected char at \r\n EOL sequence.');
                }
                ++this.m_current_char_index;
            }
            else if (10 === char_code) {
                ++this.m_current_char_index;
            }
            else {
                this.Throw('Unexpected char at EOL.');
            }
            ++this.m_current_line_index;
            this.m_current_line_first_char_index = this.m_current_char_index;
        }
    }
    Skip_SingleLineComment() {
        do {
            const current_char_code = this.m_template.charCodeAt(this.m_current_char_index);
            if ((13 === current_char_code)
                ||
                    (10 === current_char_code)) {
                break;
            }
            ++this.m_current_char_index;
        } while (this.m_template.length !== this.m_current_char_index);
    }
    Parse_TextLine() {
        ++this.m_current_char_index;
        if (this.m_template.length === this.m_current_char_index) {
            this.Throw('Unexpected EOF at text block.');
        }
        const text_chars_begin_index = this.m_current_char_index;
        do {
            const char_code = this.m_template.charCodeAt(this.m_current_char_index);
            if ((13 === char_code)
                ||
                    (10 === char_code)) {
                break;
            }
            ++this.m_current_char_index;
        } while (this.m_template.length !== this.m_current_char_index);
        if (this.m_current_char_index === text_chars_begin_index) {
            this.Throw('Unexpected EOL at text block.');
        }
        return (this.m_template.substring(text_chars_begin_index, this.m_current_char_index));
    }
    Parse_Emitter() {
        this.Flush_AccumulatedTextChunks();
        this.Prepare_To_AppendCommand();
        const emitter_name = this.Parse_EmitterName();
        const emitter_closed = (47 === emitter_name.charCodeAt(emitter_name.length - 1));
        let emitter_arguments;
        if (this.m_template.length !== this.m_current_char_index) {
            const char_code = this.m_template.charCodeAt(this.m_current_char_index);
            if (40 === char_code) {
                emitter_arguments = this.Parse_EmitterArguments();
            }
        }
        const commands_start_index = this.m_commands_sequence.length;
        if (undefined === emitter_arguments) {
            this.m_commands_sequence.push(-1);
        }
        else {
            this.m_commands_sequence.push(-1 - emitter_arguments.length);
            emitter_arguments.forEach(function (value) { this.push(value); }, this.m_commands_sequence);
        }
        this.m_commands_sequence.push(emitter_name);
        this.m_last_block_closed = emitter_closed;
        if (!emitter_closed) {
            this.m_opened_blocks_commands_starts_indexes.push(commands_start_index);
        }
        if (this.m_template.length !== this.m_current_char_index) {
            const char_code = this.m_template.charCodeAt(this.m_current_char_index);
            if (45 === char_code) {
                this.Skip_SingleLineComment();
            }
            else if (92 === char_code) {
                if (emitter_closed) {
                    this.Throw('Unexpected char, escaped text may not appear after closed emitter.');
                }
                else {
                    ++this.m_current_depth;
                    this.Prepare_To_AppendCommand();
                    this.m_text_chunks_accumulator.push(this.m_options.escape_method(this.Parse_TextLine()));
                    this.m_last_block_closed = true;
                }
            }
            else if (61 === char_code) {
                if (emitter_closed) {
                    this.Throw('Unexpected char, raw text may not appear after closed emitter.');
                }
                else if (this.m_options.raw_text_is_allowed) {
                    ++this.m_current_depth;
                    this.Prepare_To_AppendCommand();
                    this.m_text_chunks_accumulator.push(this.Parse_TextLine());
                    this.m_last_block_closed = true;
                }
                else {
                    this.Throw('Raw text processing is prohibited');
                }
            }
            else if (13 === char_code) {
            }
            else if (10 === char_code) {
            }
            else {
                this.Throw('Unexpected char');
            }
        }
    }
    Parse_EmitterName() {
        const name_chars_begin_index = this.m_current_char_index;
        for (;;) {
            ++this.m_current_char_index;
            if (this.m_template.length === this.m_current_char_index) {
                break;
            }
            const char_code = this.m_template.charCodeAt(this.m_current_char_index);
            if (!Is_NameSafe(char_code)) {
                if (47 === char_code) {
                    ++this.m_current_char_index;
                }
                break;
            }
        }
        return (this.m_template.substring(name_chars_begin_index, this.m_current_char_index));
    }
    Parse_EmitterArguments() {
        ++this.m_current_char_index;
        let emitter_arguments = [];
        for (;;) {
            this.Skip_OptionalWhitespace();
            if (this.m_template.length === this.m_current_char_index) {
                this.Throw('Unexpected EOF at emitter arguments list.');
            }
            {
                const char_code = this.m_template.charCodeAt(this.m_current_char_index);
                if (34 === char_code) {
                    emitter_arguments.push(this.Parse_EmitterStringArgument());
                }
                else if (Is_NameSafe(char_code)) {
                    emitter_arguments.push(this.Parse_EmitterArgument());
                }
                else if ((44 === char_code) || (41 === char_code)) {
                    this.Throw('Emitter argument is missing.');
                }
                else if ((13 === char_code) || (10 === char_code)) {
                    this.Throw('Unexpected EOL at emitter arguments list.');
                }
                else {
                    this.Throw('Unexpected char at emitter arguments list.');
                }
            }
            this.Skip_OptionalWhitespace();
            if (this.m_template.length === this.m_current_char_index) {
                this.Throw('Unexpected EOF at emitter arguments list.');
            }
            {
                const char_code = this.m_template.charCodeAt(this.m_current_char_index);
                if (44 === char_code) {
                    if (exports.emitter_arguments_count_max <= emitter_arguments.length) {
                        this.Throw('Emitter arguments count limit (' + exports.emitter_arguments_count_max.toString() + ') is exceeded.');
                    }
                    ++this.m_current_char_index;
                }
                else if (41 === char_code) {
                    break;
                }
                else if ((13 === char_code) || (10 === char_code)) {
                    this.Throw('Unexpected EOL at emitter arguments list.');
                }
                else {
                    this.Throw('Unexpected char at emitter arguments list.');
                }
            }
        }
        ++this.m_current_char_index;
        return (emitter_arguments);
    }
    Parse_EmitterArgument() {
        const argument_chars_begin_index = this.m_current_char_index;
        for (;;) {
            ++this.m_current_char_index;
            if (this.m_template.length === this.m_current_char_index) {
                this.Throw('Unexpected EOF at emitter argument.');
            }
            const char_code = this.m_template.charCodeAt(this.m_current_char_index);
            if ((13 === char_code) || (10 === char_code)) {
                this.Throw('Unexpected EOL at emitter argument.');
            }
            if (false === Is_NameSafe(char_code)) {
                break;
            }
        }
        const argument = this.m_template.substring(argument_chars_begin_index, this.m_current_char_index);
        if ('undefined' !== argument) {
            this.m_current_char_index = argument_chars_begin_index;
            this.Throw('Unexpected token at emitter arguments list.');
        }
        return (undefined);
    }
    Parse_EmitterStringArgument() {
        ++this.m_current_char_index;
        let parts = [];
        let at_escape_sequence = false;
        for (;;) {
            if (this.m_template.length === this.m_current_char_index) {
                this.Throw('Unexpected EOF at emitter string argument.');
            }
            const char_code = this.m_template.charCodeAt(this.m_current_char_index);
            if (at_escape_sequence) {
                if (92 === char_code) {
                    parts.push('\\');
                }
                else if (34 === char_code) {
                    parts.push('"');
                }
                else {
                    this.Throw('Unexpected char at emitter string argument escape sequence.');
                }
                at_escape_sequence = false;
            }
            else {
                if ((13 === char_code) || (10 === char_code)) {
                    this.Throw('Unexpected EOL at emitter string argument.');
                }
                else if (34 === char_code) {
                    break;
                }
                else if (92 === char_code) {
                    at_escape_sequence = true;
                }
                else {
                    parts.push(this.m_template.charAt(this.m_current_char_index));
                }
            }
            ++this.m_current_char_index;
        }
        const string_argument = parts.join('');
        ++this.m_current_char_index;
        return (string_argument);
    }
}
TCompilationContext.s_default_options = {
    escape_method: exports.Escape_HtmlEntities,
    raw_text_is_allowed: false,
    spaces_per_ident_level: 2
};
exports.Compile = function (template, options) {
    return ((new TCompilationContext(template, options)).compiled_commands);
};
function Flush_ChunksAccumulator(prerendered_commands_sequence, chunks_accumulator) {
    if (0 !== chunks_accumulator.length) {
        const combined_chunks = chunks_accumulator.join('');
        if (0 !== combined_chunks.length) {
            prerendered_commands_sequence.push(combined_chunks);
        }
        chunks_accumulator.length = 0;
    }
}
exports.Advance = function (emitters) {
    const commands_sequence = this;
    let prerendered_commands_sequence = [];
    let opened_emitters_start_indexes = [];
    let chunks_accumulator = [];
    let command_index = 0;
    while (commands_sequence.length !== command_index) {
        const command = commands_sequence[command_index];
        if ('string' === typeof command) {
            chunks_accumulator.push(command);
            command_index += 1;
        }
        else if (command < 0) {
            const emitter_commands_start_index = command_index;
            const emitter_arguments_count = -1 - command;
            const emitter_arguments_commands_begin_index = emitter_commands_start_index + 1;
            const emitter_arguments_commands_end_index = emitter_arguments_commands_begin_index + emitter_arguments_count;
            const emitter_arguments = ((0 !== emitter_arguments_count)
                ?
                    commands_sequence.slice(emitter_arguments_commands_begin_index, emitter_arguments_commands_end_index)
                :
                    undefined);
            const emitter_name_command_index = emitter_arguments_commands_end_index;
            const emitter_name = commands_sequence[emitter_name_command_index];
            const right_slash_char_code = 47;
            const emitted_block_is_opened = (right_slash_char_code !== emitter_name.charCodeAt(emitter_name.length - 1));
            let emitter = emitters[emitter_name];
            if ((undefined !== emitter) && emitter.is_pure) {
                if (emitted_block_is_opened) {
                    emitter.Emit_Prologue(chunks_accumulator, emitters, undefined, emitter_arguments);
                }
                else {
                    emitter.Emit(chunks_accumulator, emitters, undefined, emitter_arguments);
                }
            }
            else {
                Flush_ChunksAccumulator(prerendered_commands_sequence, chunks_accumulator);
                opened_emitters_start_indexes.push(prerendered_commands_sequence.length);
                prerendered_commands_sequence.push(-1 - emitter_arguments_count);
                if (0 !== emitter_arguments_count) {
                    prerendered_commands_sequence.push.apply(prerendered_commands_sequence, emitter_arguments);
                }
                prerendered_commands_sequence.push(emitter_name);
            }
            command_index += 2 + emitter_arguments_count;
        }
        else {
            const emitter_commands_start_index = command;
            const emitter_arguments_count = -1 - commands_sequence[emitter_commands_start_index];
            const emitter_arguments_commands_begin_index = emitter_commands_start_index + 1;
            const emitter_arguments_commands_end_index = emitter_arguments_commands_begin_index + emitter_arguments_count;
            const emitter_arguments = ((0 !== emitter_arguments_count)
                ?
                    commands_sequence.slice(emitter_arguments_commands_begin_index, emitter_arguments_commands_end_index)
                :
                    undefined);
            const emitter_name_command_index = emitter_arguments_commands_end_index;
            const emitter_name = commands_sequence[emitter_name_command_index];
            let emitter = emitters[emitter_name];
            if ((undefined !== emitter) && emitter.is_pure) {
                emitter.Emit_Epilogue(chunks_accumulator, emitters, undefined, emitter_arguments);
            }
            else {
                Flush_ChunksAccumulator(prerendered_commands_sequence, chunks_accumulator);
                prerendered_commands_sequence.push(opened_emitters_start_indexes.pop());
            }
            command_index += 1;
        }
    }
    Flush_ChunksAccumulator(prerendered_commands_sequence, chunks_accumulator);
    return (prerendered_commands_sequence);
};
exports.Emit = function (chunks, emitters, locals) {
    const commands_sequence = this;
    let command_index = 0;
    while (commands_sequence.length !== command_index) {
        const command = commands_sequence[command_index];
        if ('string' === typeof command) {
            chunks.push(command);
            command_index += 1;
        }
        else if (command < 0) {
            const emitter_commands_start_index = command_index;
            const emitter_arguments_count = -1 - command;
            const emitter_arguments_commands_begin_index = emitter_commands_start_index + 1;
            const emitter_arguments_commands_end_index = emitter_arguments_commands_begin_index + emitter_arguments_count;
            const emitter_arguments = ((0 !== emitter_arguments_count)
                ?
                    commands_sequence.slice(emitter_arguments_commands_begin_index, emitter_arguments_commands_end_index)
                :
                    undefined);
            const emitter_name_command_index = emitter_arguments_commands_end_index;
            const emitter_name = commands_sequence[emitter_name_command_index];
            const right_slash_char_code = 47;
            const emitted_block_is_opened = (right_slash_char_code !== emitter_name.charCodeAt(emitter_name.length - 1));
            let emitter = emitters[emitter_name];
            if (emitted_block_is_opened) {
                emitter.Emit_Prologue(chunks, emitters, locals, emitter_arguments);
            }
            else {
                emitter.Emit(chunks, emitters, locals, emitter_arguments);
            }
            command_index += 2 + emitter_arguments_count;
        }
        else {
            const emitter_commands_start_index = command;
            const emitter_arguments_count = -1 - commands_sequence[emitter_commands_start_index];
            const emitter_arguments_commands_begin_index = emitter_commands_start_index + 1;
            const emitter_arguments_commands_end_index = emitter_arguments_commands_begin_index + emitter_arguments_count;
            const emitter_arguments = ((0 !== emitter_arguments_count)
                ?
                    commands_sequence.slice(emitter_arguments_commands_begin_index, emitter_arguments_commands_end_index)
                :
                    undefined);
            const emitter_name_command_index = emitter_arguments_commands_end_index;
            const emitter_name = commands_sequence[emitter_name_command_index];
            let emitter = emitters[emitter_name];
            emitter.Emit_Epilogue(chunks, emitters, locals, emitter_arguments);
            command_index += 1;
        }
    }
};
exports.Render = function (emitters, locals) {
    let chunks = [];
    exports.Emit.apply(this, [chunks, emitters, locals]);
    return (chunks.join(''));
};
class TRenderer {
    constructor(command_sequence) {
        this.m_commands_sequence = command_sequence;
    }
    Validate(emitters) {
        exports.Validate.apply(this.m_commands_sequence, arguments);
    }
    Advance(emitters) {
        this.m_commands_sequence = exports.Advance.apply(this.m_commands_sequence, arguments);
    }
    Render(emitters, locals) {
        return (exports.Render.apply(this.m_commands_sequence, arguments));
    }
    get what() { return ('renderer emitter'); }
    get arguments_descriptions() { return (TRenderer.TRenderer_s_arguments_descriptions); }
    get is_pure() { return (this.m_commands_sequence.length <= 1); }
    Emit(chunks, emitters, locals, emitter_arguments) {
        exports.Emit.apply(this.m_commands_sequence, arguments);
    }
}
TRenderer.TRenderer_s_arguments_descriptions = [];
exports.TRenderer = TRenderer;
//# sourceMappingURL=index.js.map